#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "AVL.h"

static void recalcule_balance_aux(AVL a, int * hauteur)
{
	int hg, hd;
	if(a == NULL)
	{
		*hauteur = -1;
		return;
	}
	recalcule_balance_aux(a->gauche, &hg);
	recalcule_balance_aux(a->droite, &hd);
	a->bal = (hd - hg);
	if(hd > hg)
		*hauteur = hd;
	else
		*hauteur = hg;
	(*hauteur) ++;
}

static void recalcule_balance(AVL a)
{
	int h;
	recalcule_balance_aux(a, &h);
}

static int check_balance_aux(const AVL a, int * hauteur)
{
	int hg, hd;
	if(a == NULL)
	{
		*hauteur = -1;
		return 1;
	}
	if(! check_balance_aux(a->gauche, &hg))
		return 0;
	if(! check_balance_aux(a->droite, &hd))
		return 0;
	if(a->bal != (hd - hg))
		return 0;
	if(hd > hg)
		*hauteur = hd;
	else
		*hauteur = hg;
	(*hauteur) ++;
	return 1;
}

static int check_balance(const AVL a)
{
	int h;
	return check_balance_aux(a, &h);
}

static int abs_int(int val)
{
	if(val < 0)
		return -val;
	return val;
}

static void rotationDroite(AVL * a)
{
	int i;
	int j;
	int ng;
	int nr;
	
	AVL tmp = NULL;
	
	if(a == NULL || *a == NULL)
		return;
	if((*a)->gauche == NULL)
		return;
	
	i = (*a)->bal;
	j = (*a)->gauche->bal;
	
	tmp = (*a)->gauche->droite;
	(*a)->gauche->droite = *a;
	*a = (*a)->gauche;
	(*a)->droite->gauche = tmp;
	
	if(j >= 0)
		nr = i + 1;
	else
		nr = i - j + 1;
	if(nr > 0)
		ng = nr + j + 1;
	else
		ng = j + 1;
	
	(*a)->bal = ng;
	(*a)->droite->bal = nr;
}

static void rotationGauche(AVL * a)
{
	int i;
	int j;
	int nd;
	int nr;
	
	AVL tmp = NULL;
	
	if(a == NULL || *a == NULL)
		return;
	if((*a)->droite == NULL)
		return;
	
	i = (*a)->bal;
	j = (*a)->droite->bal;
	
	tmp = (*a)->droite->gauche;
	(*a)->droite->gauche = *a;
	*a = (*a)->droite;
	(*a)->gauche->droite = tmp;
	
	if(j <= 0)
		nr = i - 1;
	else
		nr = i - j - 1;
	if(nr < 0)
		nd = nr + j - 1;
	else
		nd = j - 1;
	
	(*a)->bal = nd;
	(*a)->gauche->bal = nr;
}

static void rotationGaucheDroite(AVL * a)
{
	if(a == NULL || *a == NULL)
		return;
	if((*a)->gauche == NULL)
		return;
	if((*a)->gauche->droite == NULL)
		return;
	rotationGauche(&((*a)->gauche));
	rotationDroite(a);
}

static void rotationDroiteGauche(AVL * a)
{
	if(a == NULL || *a == NULL)
		return;
	if((*a)->droite == NULL)
		return;
	if((*a)->droite->gauche == NULL)
		return;
	rotationDroite(&((*a)->droite));
	rotationGauche(a);
}

static void AVL_Equilibrer(AVL * a)
{
	if(a == NULL || *a == NULL)
		return;
	if(abs_int((*a)->bal) < 2)
		return;
		
	if((*a)->bal > 0)
	{
		if((*a)->droite == NULL)
			return;
		if((*a)->droite->bal >= 0)
			rotationGauche(a);
		else
			rotationDroiteGauche(a);
	}
	else
	{
		if((*a)->gauche == NULL)
			return;
		if((*a)->gauche->bal <= 0)
			rotationDroite(a);
		else
			rotationGaucheDroite(a);
	}
}

/*
Fonctions principales
*/

AVL AVL_creer_noeud(const void * valeur)
{
	AVL a = NULL;
	
	if((a = (AVL)malloc(sizeof(NoeudAVL))) == NULL)
		return NULL;
	
	a->valeur = (void *)valeur;
	a->bal = 0;
	a->gauche = NULL;
	a->droite = NULL;
	
	return a;
}

void AVL_free(AVL * a, void (*free_element)(void *))
{
	if(a != NULL && (*a) != NULL)
	{
		AVL_free(&((*a)->gauche), free_element);
		AVL_free(&((*a)->droite), free_element);
		
		free_element(&((*a)->valeur));
		free(*a);
		*a = NULL;
	}
}

static int AVL_ajouter_aux(AVL * a, const void * valeur, Comparaison cmp, int * erreur)
{
	int pousser = 0;
	
	if(a == NULL)
	{
		if(erreur != NULL)
			*erreur = ERREUR_UTILISATION;
		return 0;
	}
	if((*a) == NULL)
	{
		if(((*a) = AVL_creer_noeud(valeur)) == NULL)
		{
			if(erreur != NULL)
				*erreur = ERREUR_MEMOIRE;
			return 0;
		}
		return 1;
	}
	
	if(cmp((*a)->valeur, valeur) > 0)
		pousser = -AVL_ajouter_aux(&((*a)->gauche), valeur, cmp, erreur);
	if(cmp((*a)->valeur, valeur) < 0)
		pousser = AVL_ajouter_aux(&((*a)->droite), valeur, cmp, erreur);
	if(pousser == 0)
		return 0;
	((*a)->bal) += pousser;
	
	AVL_Equilibrer(a);
	
	return abs_int((*a)->bal);
}

int AVL_ajouter(AVL * a, const void * valeur, Comparaison cmp)
{
	int erreur = 0;
	if(AVL_trouver(*a, valeur, cmp))
		return ERREUR_DEJA_FAIT;
	AVL_ajouter_aux(a, valeur, cmp, &erreur);
	assert(check_balance(*a));
	if(erreur != 0)
		return erreur;
	return 1;
}

static void ** AVL_mini(const AVL a)
{
	if(a == NULL)
		return 0;
	if(a->gauche == NULL)
		return &(a->valeur);
	return AVL_mini(a->gauche);
}

static void ** AVL_maxi(const AVL a)
{
	if(a == NULL)
		return 0;
	if(a->droite == NULL)
		return &(a->valeur);
	return AVL_maxi(a->droite);
}

static int AVL_extraire_aux(AVL * a, const void * valeur, Comparaison cmp, AVL * res)
{
	int retenue = 0;
	int ancienne_bal = 0;
	int autre_bal = 0;
	
	if(a == NULL || *a == NULL)
		return 0;
	
	if(cmp((*a)->valeur, valeur) == 0)
	{
		void ** nouvelle_valeur = NULL;
		void * tmp = NULL;
		if((*a)->gauche == NULL || (*a)->droite == NULL)
		{
			*res = *a;
			if((*a)->gauche == NULL)
				*a = (*a)->droite;
			else
				*a = (*a)->gauche;
			(*res)->droite = NULL;
			(*res)->gauche = NULL;
			return 1;
		}
		
		if((*a)->bal > 0)
		{
			nouvelle_valeur = AVL_mini((*a)->droite);
			retenue = AVL_extraire_aux(&((*a)->droite), *nouvelle_valeur, cmp, res);
		}
		else
		{
			nouvelle_valeur = AVL_maxi((*a)->gauche);
			retenue = -AVL_extraire_aux(&((*a)->gauche), *nouvelle_valeur, cmp, res);
		}
		tmp = (*a)->valeur;
		(*a)->valeur = *nouvelle_valeur;
		*nouvelle_valeur = tmp;
		
	}
	else if(cmp((*a)->valeur, valeur) < 0)
		retenue = AVL_extraire_aux(&((*a)->droite), valeur, cmp, res);
	else
		retenue = -AVL_extraire_aux(&((*a)->gauche), valeur, cmp, res);
	
	if(retenue == 0)
		return 0;
	
	((*a)->bal) -= retenue;
	ancienne_bal = (*a)->bal;
	
	if((*a)->bal == -2)
		autre_bal = (*a)->gauche->bal;
	else if((*a)->bal == 2)
		autre_bal = (*a)->droite->bal;
	else if((*a)->bal == 0)
		return 1;
	else
		return 0;
	
	AVL_Equilibrer(a);
	
	if(ancienne_bal == -2)
	{
		if(autre_bal == 0)
		{
			(*a)->bal = 1;
			(*a)->droite->bal = -1;
			retenue = 0;
		}
		else
		{
			/*(*a)->bal = 0;
			(*a)->droite->bal = 0;*/
			recalcule_balance(*a);
		}
	}
	else if(ancienne_bal == 2)
	{
		if(autre_bal == 0)
		{
			(*a)->bal = -1;
			(*a)->gauche->bal = 1;
			retenue = 0;
		}
		else
		{
			/*(*a)->bal = 0;
			(*a)->gauche->bal = 0;*/
			recalcule_balance(*a);
		}
	}
	
	return abs_int(retenue);
}

AVL AVL_extraire(AVL * a, const void * valeur, Comparaison cmp)
{
	AVL res = NULL;
	if(! AVL_trouver(*a, valeur, cmp))
		return NULL;
	
	AVL_extraire_aux(a, valeur, cmp, &res);
	assert(check_balance(*a));
	return res;
}

AVL AVL_trouver(const AVL a, const void * valeur, Comparaison cmp)
{
	if(a == NULL)
		return NULL;
	if(cmp(a->valeur, valeur) == 0)
		return a;
	else if(cmp(a->valeur, valeur) > 0)
		return AVL_trouver(a->gauche, valeur, cmp);
	else
		return AVL_trouver(a->droite, valeur, cmp);
}

AVL AVL_trouver_split(const AVL a, const void * valeur, Comparaison cmp)
{
	if(a == NULL)
		return NULL;
	if(cmp(a->valeur, valeur) == 0)
		return a;
	else if(cmp(a->valeur, valeur) == SPLIT_VALUE)
	{
		AVL vg = NULL, vd = NULL;
		if((vg = AVL_trouver(a->gauche, valeur, cmp)) == NULL)
			return vd = AVL_trouver(a->droite, valeur, cmp);
		else
			return vg;
	}
	else if(cmp(a->valeur, valeur) > 0)
		return AVL_trouver(a->gauche, valeur, cmp);
	else
		return AVL_trouver(a->droite, valeur, cmp);
}
