#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "automate_aff_debug.h"

void AVL_afficher_debug(const AVL cur, void (*elem_aff)(const void *), int tab)
{
	int i;
	if(cur == NULL)
		return;
	AVL_afficher_debug(cur->gauche, elem_aff, tab + 1);
	for(i = 0; i < tab; ++i)putchar('.');
	if(elem_aff != NULL)
		(*elem_aff)(cur->valeur);
	printf("| %d\n", cur->bal);
	
	AVL_afficher_debug(cur->droite, elem_aff, tab + 1);
}

void Automate_afficher(Automate automate)
{
	printf("Alphabet:");
	Ensemble_afficher((void *)(automate->alphabet));
	putchar('\n');
	printf("Etats:");
	Ensemble_afficher((void *)(automate->etats));
	putchar('\n');
	printf("Etats Initiaux:");
	Ensemble_afficher((void *)(automate->etats_init));
	putchar('\n');
	printf("Etats Finaux:");
	Ensemble_afficher((void *)(automate->etats_finaux));
	putchar('\n');
	printf("Transitions:");
	Ensemble_afficher((void *)(automate->transitions));
	putchar('\n');
}

static void Ensemble_afficher_aux(const AVL abr, void (*aff)(const void *))
{
	if(abr == NULL)
		return;
	Ensemble_afficher_aux(abr->gauche, aff);
	(*aff)(abr->valeur);
	putchar('\n');
	Ensemble_afficher_aux(abr->droite, aff);
}

void Ensemble_afficher(const void * ens)
{
	if(ens == NULL)
		return;
	putchar('{');
	/*AVL_afficher_debug(((Ensemble)ens)->elements, Type_afficher(((Ensemble)ens)->type, ((Ensemble)ens)->liens), 0);*/
	Ensemble_afficher_aux(((Ensemble)ens)->elements, Type_afficher(((Ensemble)ens)->type, ((Ensemble)ens)->liens));
	putchar('}');
}

void (*Type_afficher(Element_Type type, int lien))(const void *)
{
	switch(type)
	{
		case TYPE_ENTIER:
		return &int_afficher;
		case TYPE_MOT:
		return &Mot_afficher;
		case TYPE_ETAT:
		/*if(lien)*/
			return &Etat_afficher;
		/*else
			return &Etat_afficher_detail;*/
		case TYPE_TRANSITION:
		return &Transition_afficher;
	}
	return &ptr_afficher;
}

void Etat_afficher(const void * etat)
{
	if(etat == NULL)
		return;
	putchar('(');
	Mot_afficher((void *)((Etat)etat)->nom);
	putchar(')');
}

void Etat_afficher_detail(const void * etat)
{
	if(etat == NULL)
		return;
	Etat_afficher(etat);
	Ensemble_afficher((void *)((Etat)etat)->suivants);
	Ensemble_afficher((void *)((Etat)etat)->precedents);
}

void Transition_afficher(const void * transition)
{
	if(transition == NULL)
		return;
	Etat_afficher((void *)((Transition)transition)->depart);
	printf("--");
	Mot_afficher((void *)((Transition)transition)->lettre);
	printf("->");
	Etat_afficher((void *)((Transition)transition)->arrivee);
}

void Transition_afficher_debug(const void * transition)
{
	if(transition == NULL)
		return;
	Etat_afficher((void *)((Transition)transition)->depart);
	printf("[%p]", (void *)((Transition)transition)->depart);
	printf("--");
	Mot_afficher((void *)((Transition)transition)->lettre);
	printf("->");
	Etat_afficher((void *)((Transition)transition)->arrivee);
	printf("[%p]", (void *)((Transition)transition)->arrivee);
}

void Mot_afficher(const void * mot)
{
	if(mot == NULL)
		return;
	printf("%s", (char *)mot);
}

void int_afficher(const void * entier)
{
	if(entier == NULL)
		return;
	printf("%d", *(int *)entier);
}

void ptr_afficher(const void * ptr)
{
	printf("%p", ptr);
}

Mot (*Element_chaine(Element_Type type))(const void *)
{
	switch(type)
	{
		case TYPE_TRANSITION:
		return &Transition_chaine;
		case TYPE_ETAT:
		return &Etat_chaine;
		case TYPE_MOT:
		return &Mot_chaine;
		case TYPE_ENTIER:
		return &int_chaine;
	}
	return &ptr_chaine;
}

static int Ensemble_chaine_aux(const AVL cur, Mot * liste, int * indice, Mot (*chaine)(const void *))
{
	if(cur == NULL)
		return 1;
	if(! Ensemble_chaine_aux(cur->gauche, liste, indice, chaine))
		return 0;
	if((*(liste + *indice) = (*chaine)(cur->valeur)) == NULL)
		return 0;
	++(*indice);
	if(! Ensemble_chaine_aux(cur->droite, liste, indice, chaine))
		return 0;
	return 1;
}

Mot Ensemble_chaine(const void * element)
{
	Ensemble ens = NULL;
	Mot * liste = NULL;
	Mot mot = NULL;
	int i = 0;
	int nb_elements = 0;
	
	if((ens = (Ensemble)element) == NULL)
		return NULL;
	nb_elements = Ensemble_nombre_elements(ens);
	if((liste = (Mot *)calloc(nb_elements, sizeof(Mot))) == NULL)
		return NULL;
	if(Ensemble_chaine_aux(ens->elements, liste, &i, Element_chaine(ens->type)))
	{
		int taille = 2;
		for(i = 0; i < nb_elements; ++i)
		{
			if(i != 0)
				taille ++;
			taille += strlen((char *)(*(liste + i)));
		}
		if((mot = (Mot)malloc((taille + 1) * sizeof(Mot_aux))) != NULL)
		{
			strcpy((char *)mot, "{");
			for(i = 0; i < nb_elements; ++i)
			{
				if(i != 0)
					strcat((char *)mot, ",");
				strcat((char *)mot, (char *)*(liste + i));
			}
			strcat((char *)mot, "}");
		}
	}
	for(i = 0; i < nb_elements; ++i)
		Element_free(TYPE_MOT)((liste + i));
	free(liste);
	return mot;
}

Mot Transition_chaine(const void * element)
{
	Transition transition = NULL;
	Mot mot = NULL;
	Mot etat_depart = NULL, etat_arrivee = NULL, lettre = NULL;;
	if((transition = (Transition)element) == NULL)
		return NULL;
	if((etat_depart = Etat_chaine(transition->depart)) == NULL
	|| (etat_arrivee = Etat_chaine(transition->arrivee)) == NULL
	|| (lettre = Mot_chaine(transition->lettre)) == NULL
	|| (mot = (Mot)malloc((strlen((char *)etat_depart) + strlen((char *)etat_arrivee) + strlen((char *)lettre) + 4) * sizeof(Mot_aux))) != NULL)
	{
		sprintf("%s-%s->%s", (char *)etat_depart, (char *)lettre, (char *)etat_arrivee);
	}
	
	Element_free(TYPE_MOT)(&etat_depart);
	Element_free(TYPE_MOT)(&etat_arrivee);
	Element_free(TYPE_MOT)(&lettre);
	return mot;
}

Mot Etat_chaine(const void * element)
{
	Etat etat = NULL;
	if((etat = (Etat)element) == NULL)
		return NULL;
	return Mot_chaine(etat->nom);
}

Mot Mot_chaine(const void * mot)
{
	return Mot_creer((Mot)mot);
}

Mot int_chaine(const void * entier)
{
	Mot mot = NULL;
	int taille = 1;
	int nb;
	
	if(entier == NULL)
		return NULL;
		
	nb = *((int *)entier);
	if(nb > 0)
		taille = (int)(log10((double)nb)) + 1;
	else if(nb < 0)
		taille = (int)(log10((double)nb)) + 2;
	if((mot = (Mot)malloc((taille + 1) * sizeof(Mot_aux))) == NULL)
		return NULL;
	sprintf((char *)mot, "%d", *((int *)entier));
	return mot;
}

Mot ptr_chaine(const void * ptr)
{
	Mot mot = NULL;
	int taille = 5;/*(nil)*/
	double nb = (double)((unsigned long int)ptr);
	if(nb > 0)
		taille = (int)(log(nb)/log(16)) + 3;
	if((mot = (Mot)malloc((taille + 1) * sizeof(Mot_aux))) == NULL)
		return NULL;
	sprintf((char *)mot, "%p", ptr);
	return mot;
}

Mot couple_chaine(const void * prem, const void * sec, Mot (*prem_chaine)(const void *), Mot (*sec_chaine)(const void *))
{
	Mot prem_mot = NULL;
	Mot sec_mot = NULL;
	Mot res = NULL;
	
	if((prem_mot = (*prem_chaine)(prem)) == NULL)
		return NULL;
	if((sec_mot = (*sec_chaine)(sec)) == NULL)
		return NULL;
	if((res = (Mot)malloc((strlen((char *)prem_mot) + strlen((char *)sec_mot) + 4) * sizeof(Mot_aux))) == NULL)
		return NULL;
	sprintf((char *)res, "(%s|%s)", (char *)prem_mot, (char *)sec_mot);
	Element_free(TYPE_MOT)(&prem_mot);
	Element_free(TYPE_MOT)(&sec_mot);
	return res;
}
