#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "automate_edit.h"
#include "automate_propriete.h"
#include "automate_aff_debug.h"
#include "AVL.h"

static int est_special_car(Mot lettre)
{
	return ((*lettre) & 128) == 128;
}

static int est_special_car_suite(Mot lettre)
{
	return est_special_car(lettre) && (((~(*lettre)) & 64) == 64);
}

static int taille_lettre(Mot mot)
{
	int taille;
	if(mot == NULL)
		return 0;
	if(*mot == '\0')
		return 0;
	
	if(! est_special_car(mot))
		return 1;
	++mot;
	for(taille = 1; est_special_car_suite(mot); ++mot, ++taille);
	return taille;
}

Mot premiere_lettre(Mot mot)
{
	Mot res = NULL;
	int taille = 1; /*Gérer UTF8 plus tard*/
	int i;
	taille = taille_lettre(mot);
	if((res = (Mot)malloc((taille + 1) * sizeof(Mot_aux))) == NULL)
		return NULL;
	for(i = 0; i < taille; ++i)
		*(res + i) = *(mot + i);
	*(res + taille) = '\0';
	return res;
}

int Automate_ajouter_lettre(Automate automate, Mot lettre)
{
	Mot valeur = NULL;
	
	if(automate == NULL)
		return 0;
	if((valeur = premiere_lettre(lettre)) == NULL)
		return 0;
	if(cmp_mot(lettre, valeur) != 0
	|| Automate_existe_lettre(automate, valeur)
	|| ! Ensemble_ajouter(automate->alphabet, (void *)valeur))
	{
		Element_free(TYPE_MOT)(&valeur);
		return 0;
	}
	return 1;
}

int Automate_ajouter_etat(Automate automate, Mot etat_nom)
{
	Etat etat = NULL;
	
	if(automate == NULL)
		return 0;
	if(Automate_existe_etat(automate, etat_nom))
		return 0;
	if((etat = Etat_creer(etat_nom)) == NULL)
		return 0;
		
	if(! Ensemble_ajouter(automate->etats, (void *)etat))
	{
		Element_free(TYPE_ETAT)(&etat);
		return 0;
	}
	return 1;
}

int Automate_ajouter_etat_initial(Automate automate, Mot etat_nom)
{
	Etat etat = NULL;
	
	if((etat = Ensemble_trouver_etat_nom(automate->etats, etat_nom)) == NULL)
		return 0;
		
	if(! Ensemble_ajouter(automate->etats_init, (void *)etat))
		return 0;
	
	return 1;
}

int Automate_ajouter_etat_final(Automate automate, Mot etat_nom)
{
	Etat etat = NULL;
	
	if((etat = Ensemble_trouver_etat_nom(automate->etats, etat_nom)) == NULL)
		return 0;
		
	if(! Ensemble_ajouter(automate->etats_finaux, (void *)etat))
		return 0;
	
	return 1;
}

int Automate_ajouter_transition(Automate automate, Mot etat_depart, Mot lettre, Mot etat_arrivee)
{
	Transition transition = NULL;
	Etat depart = NULL, arrivee = NULL;
	
	if(automate == NULL)
		return 0;
	if(Automate_existe_transition(automate, etat_depart, lettre, etat_arrivee))
		return 0;
		
	if((depart = Ensemble_trouver_etat_nom(automate->etats, etat_depart)) == NULL)
		return 0;
	if((arrivee = Ensemble_trouver_etat_nom(automate->etats, etat_arrivee)) == NULL)
		return 0;
	
	if(! Automate_existe_lettre(automate, lettre))
	{
		if(! Automate_ajouter_lettre(automate, lettre))
			return 0;
	}
		
	if((transition = Transition_creer(depart, lettre, arrivee)) == NULL)
		return 0;
	
	if(! Ensemble_ajouter(automate->transitions, (void *)transition)
	|| ! Ensemble_ajouter(depart->suivants, (void *)transition)
	|| ! Ensemble_ajouter(arrivee->precedents, (void *)transition))
	{
		Ensemble_supprimer(automate->transitions, (void *)transition);
		Ensemble_supprimer(depart->suivants, (void *)transition);
		Ensemble_supprimer(arrivee->precedents, (void *)transition);
		Element_free(TYPE_TRANSITION)(&transition);
		return 0;
	}
	return 1;
}

static int Automate_ajouter_transitions(Automate dest, const AVL cur)
{
	Transition trans = NULL;
	if(cur == NULL)
		return 1;
	trans = (Transition)(cur->valeur);
	if(! Automate_existe_transition(dest, trans->depart->nom, trans->lettre, trans->arrivee->nom))
	{
		if(! Automate_ajouter_transition(dest, trans->depart->nom, trans->lettre, trans->arrivee->nom))
			return 0;
	}
	if(! Automate_ajouter_transitions(dest, cur->gauche))
		return 0;
	if(! Automate_ajouter_transitions(dest, cur->droite))
		return 0;
	return 1;
}

static int Automate_ajouter_etats(Automate dest, const AVL cur, int (*ajout)(Automate, Mot))
{
	if(cur == NULL)
		return 1;
	
	if(ajout != &Automate_ajouter_etat || ! Automate_existe_etat(dest, ((Etat)(cur->valeur))->nom))
	{
		if(! (*ajout)(dest, ((Etat)(cur->valeur))->nom))
			return 0;
	}
	if(! Automate_ajouter_etats(dest, cur->gauche, ajout))
		return 0;
	if(! Automate_ajouter_etats(dest, cur->droite, ajout))
		return 0;
	return 1;
}

int Automate_ajouter_automate(Automate automate, Automate autre, Mot prefixe1, Mot prefixe2)
{
	Automate ajout = NULL;
	if(automate == NULL || autre == NULL)
		return 0;
	if(! Automate_copier(&ajout, autre))
		return 0;
	if(! Automate_renommer_etats_par_prefixe(automate, prefixe1)
	|| ! Automate_renommer_etats_par_prefixe(ajout, prefixe2)
	|| ! Ensemble_union(automate->alphabet, ajout->alphabet)
	|| ! Automate_ajouter_etats(automate, ajout->etats->elements, Automate_ajouter_etat)
	|| ! Automate_ajouter_etats(automate, ajout->etats_init->elements, Automate_ajouter_etat_initial)
	|| ! Automate_ajouter_etats(automate, ajout->etats_finaux->elements, Automate_ajouter_etat_final)
	|| ! Automate_ajouter_transitions(automate, ajout->transitions->elements))
	{
		Automate_free(&ajout);
		return 0;
	}
	Automate_free(&ajout);
	return 1;
}

static int Automate_copier_transitions(Automate dest, const AVL cur)
{
	Transition trans = NULL;
	if(cur == NULL)
		return 1;
	trans = (Transition)(cur->valeur);
	if(! Automate_ajouter_transition(dest, trans->depart->nom, trans->lettre, trans->arrivee->nom))
		return 0;
	if(! Automate_copier_transitions(dest, cur->gauche))
		return 0;
	if(! Automate_copier_transitions(dest, cur->droite))
		return 0;
	return 1;
}

static int Automate_copier_etats(Automate dest, const AVL cur, int (*ajout)(Automate, Mot))
{
	if(cur == NULL)
		return 1;
	
	if(! (*ajout)(dest, ((Etat)(cur->valeur))->nom))
		return 0;
	if(! Automate_copier_etats(dest, cur->gauche, ajout))
		return 0;
	if(! Automate_copier_etats(dest, cur->droite, ajout))
		return 0;
	return 1;
}

int Automate_copier(Automate * dest, Automate source)
{
	Automate nouv = NULL;
	if((nouv = Automate_creer()) == NULL)
		return 0;
	if(! Ensemble_union(nouv->alphabet, source->alphabet)
	|| ! Automate_copier_etats(nouv, source->etats->elements, Automate_ajouter_etat)
	|| ! Automate_copier_etats(nouv, source->etats_init->elements, Automate_ajouter_etat_initial)
	|| ! Automate_copier_etats(nouv, source->etats_finaux->elements, Automate_ajouter_etat_final)
	|| ! Automate_copier_transitions(nouv, source->transitions->elements))
	{
		Automate_free(&nouv);
		return 0;
	}
	*dest = nouv;
	return 1;
}

int Automate_existe_lettre(Automate automate, Mot lettre)
{
	if(automate == NULL)
		return 0;
	return Ensemble_trouver(automate->alphabet, (void *)lettre);
}

int Automate_existe_etat(Automate automate, Mot etat_nom)
{
	if(automate == NULL)
		return 0;
	return Ensemble_trouver_etat_nom(automate->etats, etat_nom) != NULL;
}

int Automate_existe_transition(Automate automate, Mot etat_depart, Mot lettre, Mot etat_arrivee)
{
	Etat depart = NULL, arrivee = NULL;
	
	if(automate == NULL)
		return 0;
	if((depart = Ensemble_trouver_etat_nom(automate->etats, etat_depart)) == NULL)
		return 0;
	if((arrivee = Ensemble_trouver_etat_nom(automate->etats, etat_arrivee)) == NULL)
		return 0;
	return Ensemble_trouver_transition(automate->transitions, depart, lettre, arrivee) != NULL;
}

int Automate_supprimer_lettre(Automate automate, Mot lettre)
{
	if(automate == NULL)
		return 0;
	if(! Automate_existe_lettre(automate, lettre))
		return 0;
	if(! Automate_lettre_sans_transition(automate, lettre))
		return 0;
	
	return Ensemble_supprimer(automate->alphabet, (void *)lettre);
}

int Automate_supprimer_etat(Automate automate, Mot etat_nom)
{
	Etat etat = NULL;
	if(automate == NULL)
		return 0;
	if(! Automate_existe_etat(automate, etat_nom))
		return 0;
	if(! Automate_etat_isole(automate, etat_nom))
		return 0;
	if((etat = Ensemble_trouver_etat_nom(automate->etats, etat_nom)) == NULL)
		return 0;
	
	Ensemble_supprimer(automate->etats_init, (void *)etat);
	Ensemble_supprimer(automate->etats_finaux, (void *)etat);
	return Ensemble_supprimer(automate->etats, (void *)etat);
}

int Automate_supprimer_transition(Automate automate, Mot etat_depart, Mot lettre, Mot etat_arrivee)
{
	Transition transition = NULL;
	Etat depart = NULL, arrivee = NULL;
	if(automate == NULL)
		return 0;
		
	if((depart = Ensemble_trouver_etat_nom(automate->etats, etat_depart)) == NULL)
		return 0;
		
	if((arrivee = Ensemble_trouver_etat_nom(automate->etats, etat_arrivee)) == NULL)
		return 0;
	if((transition = Ensemble_trouver_transition(automate->transitions, depart, lettre, arrivee)) == NULL)
		return 0;
	
	Ensemble_supprimer(depart->suivants, transition);
	Ensemble_supprimer(arrivee->precedents, transition);
	if(! Ensemble_supprimer(automate->transitions, transition))
		return 0;
	return 1;
}

static int Automate_supprimer_lettres_sans_transitions_aux(Automate automate, const AVL abr)
{
	if(abr == NULL)
		return 1;
	if(Automate_lettre_sans_transition(automate, (Mot)(abr->valeur)))
	{
		if(! Automate_supprimer_lettre(automate, (Mot)(abr->valeur)))
			return 0;
	}
	if(! Automate_supprimer_lettres_sans_transitions_aux(automate, abr->gauche))
		return 0;
	if(! Automate_supprimer_lettres_sans_transitions_aux(automate, abr->droite))
		return 0;
	return 1;
}

int Automate_supprimer_lettres_sans_transitions(Automate automate)
{
	Ensemble lettres = NULL;
	int retour = 0;
	
	if(automate == NULL)
		return 0;
	
	if(! Ensemble_copier(&lettres, automate->alphabet, 0))
		return 0;
	
	retour = Automate_supprimer_lettres_sans_transitions_aux(automate, lettres->elements);
	
	Ensemble_free(&lettres);
	
	return retour;
}

static int Automate_supprimer_transitions_par_AVL(Automate automate, const AVL cur)
{
	Transition trans = NULL;
	if(cur == NULL)
		return 1;
		
	if((trans = (Transition)(cur->valeur)) == NULL)
		return 0;
		
	Automate_supprimer_transition(automate, trans->depart->nom, trans->lettre, trans->arrivee->nom);
	Automate_supprimer_transitions_par_AVL(automate, cur->gauche);
	Automate_supprimer_transitions_par_AVL(automate, cur->droite);
	return 1;
}

int Automate_supprimer_transitions_etat(Automate automate, Mot etat_nom)
{
	Ensemble transitions = NULL;
	Etat etat = NULL;
	
	if((etat = Ensemble_trouver_etat_nom(automate->etats, etat_nom)) == NULL)
		return 0;
		
	if(! Ensemble_copier(&transitions, etat->suivants, 0))
		return 0;
		
	if(! Ensemble_union(transitions, etat->precedents))
	{
		Ensemble_free(&transitions);
		return 0;
	}
	
	Automate_supprimer_transitions_par_AVL(automate, transitions->elements);
	Ensemble_free(&transitions);
	
	return 1;
}

static int Automate_supprimer_etats_prop(Automate automate, const AVL cur, int (*methode)(Automate, Mot))
{
	if(cur == NULL)
		return 1;
		
	(*methode)(automate, ((Etat)(cur->valeur))->nom);
	if(! Automate_supprimer_etats_prop(automate, cur->gauche, methode)
	|| ! Automate_supprimer_etats_prop(automate, cur->droite, methode))
		return 0;
	return 1;
}

int Automate_supprimer_etats_ensemble(Automate automate, Ensemble etats)
{
	if(automate == NULL || etats == NULL)
		return 0;
	return Automate_supprimer_etats_prop(automate, etats->elements, Automate_supprimer_etat);
}

int Automate_supprimer_transitions_etats_ensemble(Automate automate, Ensemble etats)
{
	if(automate == NULL || etats == NULL)
		return 0;
	return Automate_supprimer_etats_prop(automate, etats->elements, Automate_supprimer_transitions_etat);
}

static int Automate_etats_generer_ensemble(Automate automate, Ensemble etats, const AVL cur, int (*prop)(Automate, Mot))
{
	if(cur == NULL)
		return 1;
	if((*prop)(automate, ((Etat)(cur->valeur))->nom))
	{
		if(! Ensemble_ajouter(etats, cur->valeur))
			return 0;
	}
	if(! Automate_etats_generer_ensemble(automate, etats, cur->gauche, prop))
		return 0;
	if(! Automate_etats_generer_ensemble(automate, etats, cur->droite, prop))
		return 0;
	return 1;	
}

static int Automate_supprimer_etats_propriete(Automate automate, int (*prop)(Automate, Mot))
{
	Ensemble etats = NULL;
	
	if(automate == NULL)
		return 0;
		
	if((etats = Ensemble_creer(TYPE_ETAT, 1)) == NULL)
		return 0;
	if(! Automate_etats_generer_ensemble(automate, etats, automate->etats->elements, prop))
		return 0;
		
	Automate_supprimer_transitions_etats_ensemble(automate, etats);
	Automate_supprimer_etats_ensemble(automate, etats);
	
	Ensemble_free(&etats);
	
	return 1;
}

int Automate_supprimer_etats_non_accessibles(Automate automate)
{
	return Automate_supprimer_etats_propriete(automate, Automate_etat_est_non_accessible);
}

int Automate_supprimer_etats_non_coaccessibles(Automate automate)
{
	return Automate_supprimer_etats_propriete(automate, Automate_etat_est_non_coaccessible);
}

int Automate_lettre_sans_transition(Automate automate, Mot lettre)
{
	if(automate == NULL)
		return 1;
	return ! Ensemble_trouver_par(automate->transitions, (void *)lettre, cmp_transition_et_lettre);
}

int Automate_etat_isole(Automate automate, Mot etat_nom)
{
	Etat etat = NULL;
	if(automate == NULL)
		return 1;
	if((etat = Ensemble_trouver_etat_nom(automate->etats, etat_nom)) == NULL)
		return 1;
		
	if(Ensemble_trouver_par(automate->transitions, (void *)etat, cmp_transition_et_depart))
		return 0;
	if(Ensemble_trouver_par(automate->transitions, (void *)etat, cmp_transition_et_arrivee))
		return 0;
	return 1;
}

static int Automate_renommer_etat_par_prefixe(const AVL cur, Mot prefixe)
{
	Mot nouv = NULL;
	if(cur == NULL)
		return 1;
	if((nouv = (Mot)malloc((strlen((char *)prefixe) + strlen((char *)(((Etat)(cur->valeur))->nom)) + 1) * sizeof(Mot_aux))) == NULL)
		return 0;
	sprintf((char *)nouv, "%s%s", (char *)prefixe, (char *)(((Etat)(cur->valeur))->nom));
	Element_free(TYPE_MOT)(&(((Etat)(cur->valeur))->nom));
	((Etat)(cur->valeur))->nom = nouv;
	if(! Automate_renommer_etat_par_prefixe(cur->gauche, prefixe))
		return 0;
	if(! Automate_renommer_etat_par_prefixe(cur->droite, prefixe))
		return 0;
	return 1;
}

int Automate_renommer_etats_par_prefixe(Automate automate, Mot prefixe)
{
	if(automate == NULL)
		return 0;
	return Automate_renommer_etat_par_prefixe(automate->etats->elements, prefixe);
	
}

static int Automate_renommer_etat_par_nombre(Automate automate, const AVL cur, int * nombre)
{
	AVL etat = NULL;
	AVL etat_init = NULL;
	AVL etat_final = NULL;
	int est_init = 0;
	int est_final = 0;
	Etat val = NULL;
	Mot nouv_nom = NULL;
	
	if(cur == NULL)
		return 1;
	val = (Etat)(cur->valeur);
	if((nouv_nom = int_chaine(nombre)) == NULL)
		return 0;
	++(*nombre);
	
	if((etat = AVL_extraire(&(automate->etats->elements), val, cmp_etat)) == NULL)
	{
		Element_free(TYPE_MOT)(&nouv_nom);
		return 0;
	}
	est_init = (etat_init = AVL_extraire(&(automate->etats_init->elements), val, cmp_etat)) != NULL;
	est_final = (etat_final = AVL_extraire(&(automate->etats_finaux->elements), val, cmp_etat)) != NULL;
	
	Element_free(TYPE_MOT)(&((val->nom)));
	val->nom = nouv_nom;
	AVL_free(&etat, &Void_free);
	AVL_free(&etat_init, &Void_free);
	AVL_free(&etat_final, &Void_free);
	
	if(! AVL_ajouter(&(automate->etats->elements), val, cmp_etat))
		return 0;
	if(est_init)
	{
		if(! AVL_ajouter(&(automate->etats_init->elements), val, cmp_etat))
			return 0;
	}
	if(est_final)
	{
		if(! AVL_ajouter(&(automate->etats_finaux->elements), val, cmp_etat))
			return 0;
	}
	
	if(! Automate_renommer_etat_par_nombre(automate, cur->gauche, nombre))
		return 0;
	if(! Automate_renommer_etat_par_nombre(automate, cur->droite, nombre))
		return 0;
	
	return 1;
}

/*static int remplire_etat_liste(Etat * liste, const AVL cur, int * id)
{
	if(cur == NULL)
		return 1;
	if(! remplire_etat_liste(liste, cur->gauche, id))
		return 0;
	if((liste[*id] = (Etat)(cur->valeur)) == NULL)
		return 0;
	++(*id);
	if(! remplire_etat_liste(liste, cur->droite, id))
		return 0;
	return 1;
}

static int id_par_etat_liste(Etat * liste, Etat cur, int maxi)
{
	int i;
	for(i = 0; i < maxi; ++i)
	{
		if(cmp_etat(*(liste + i), cur) == 0)
			return i;
	}
	return -1;
}

static int Automate_renommer_par_nombre_creer_etat(Automate automate, Automate nouv, Etat * liste, int id, int maxi);

static int Automate_renommer_par_nombre_transitions(Automate automate, Automate nouv, Etat * liste, int maxi, Mot nom_etat, const AVL cur)
{
	Transition trans = NULL;
	Mot nom_arrivee = NULL;
	int id_arrivee;
	
	if(cur == NULL)
		return 1;
	if((trans = (Transition)(cur->valeur)) == NULL)
		return 0;
	if((id_arrivee = id_par_etat_liste(liste, trans->arrivee, maxi)) == -1)
		return 0;
	if((nom_arrivee = int_chaine(&id_arrivee)) == NULL)
		return 0;
	if(! Automate_existe_etat(nouv, nom_arrivee))
	{
		if(! Automate_renommer_par_nombre_creer_etat(automate, nouv, liste, id_arrivee, maxi))
		{
			Element_free(TYPE_MOT)(&nom_arrivee);
			return 0;
		}
	}
	if(! Automate_existe_transition(nouv, nom_etat, trans->lettre, nom_arrivee))
	{
		if(! Automate_ajouter_transition(nouv, nom_etat, trans->lettre, nom_arrivee))
		{
			Element_free(TYPE_MOT)(&nom_arrivee);
			return 0;
		}
	}
	Element_free(TYPE_MOT)(&nom_arrivee);
	if(! Automate_renommer_par_nombre_transitions(automate, nouv, liste, maxi, nom_etat, cur->gauche))
		return 0;
	if(! Automate_renommer_par_nombre_transitions(automate, nouv, liste, maxi, nom_etat, cur->droite))
		return 0;
	return 1;
}

static int Automate_renommer_par_nombre_creer_etat(Automate automate, Automate nouv, Etat * liste, int id, int maxi)
{
	Mot nom = NULL;
	if((nom = int_chaine(&id)) == NULL)
		return 0;
	if(Automate_existe_etat(nouv, nom))
	{
		Element_free(TYPE_MOT)(&nom);
		return 1;
	}
	if(! Automate_ajouter_etat(nouv, nom))
		return 0;
	if(Ensemble_trouver(automate->etats_init, liste[id]))
	{
		if(! Automate_ajouter_etat_initial(nouv, nom))
			return 0;
	}
	if(Ensemble_trouver(automate->etats_finaux, liste[id]))
	{
		if(! Automate_ajouter_etat_final(nouv, nom))
			return 0;
	}
	if(! Automate_renommer_par_nombre_transitions(automate, nouv, liste, maxi, nom, liste[id]->suivants->elements))
	{
		Element_free(TYPE_MOT)(&nom);
		return 0;
	}
	Element_free(TYPE_MOT)(&nom);
	return 1;
}*/

int Automate_restructure_etat_transitions(const AVL cur)
{
	Etat etat = NULL;
	if(cur == NULL)
		return 1;
	etat = (Etat)(cur->valeur);
	if(! Ensemble_restructurer(etat->suivants)
	|| ! Ensemble_restructurer(etat->precedents)
	|| ! Automate_restructure_etat_transitions(cur->gauche)
	|| ! Automate_restructure_etat_transitions(cur->droite))
		return 0;
	return 1;
}

int Automate_renommer_etats_par_nombre(Automate automate)
{
	/*int i = 0;
	int nb_elements = 0;
	Automate nouv = NULL;
	Etat * liste = NULL;
	
	if(automate == NULL || *automate == NULL)
		return 0;
	if((nouv = Automate_creer()) == NULL)
		return 0;
	if((nb_elements = Ensemble_nombre_elements((*automate)->etats)) <= 0)
		return 0;
	if((liste = (Etat *)calloc(nb_elements, sizeof(Etat))) == NULL)
	{
		Automate_free(&nouv);
		return 0;
	}
	if(! remplire_etat_liste(liste, (*automate)->etats->elements, &i))
	{
		free(liste);
		Automate_free(&nouv);
		return 0;
	}
	for(i = 0; i < nb_elements; ++i)
	{
		if(! Automate_renommer_par_nombre_creer_etat(*automate, nouv, liste, i, nb_elements))
		{
			free(liste);
			Automate_free(&nouv);
			return 0;
		}
	}
	free(liste);
	Automate_free(automate);
	*automate = nouv;
	return 1;*/
	int nombre = 0;
	int retour;
	Ensemble etats = NULL;
	
	if(automate == NULL)
		return 0;
		
	if(! Ensemble_copier(&etats, automate->etats, 1))
		return 0;
	retour = Automate_renommer_etat_par_nombre(automate, etats->elements, &nombre);
	retour &= Ensemble_restructurer(automate->transitions) && Automate_restructure_etat_transitions(etats->elements);
	Ensemble_free(&etats);
	return retour;
}
