#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "automate_all.h"
#include "automate_lexique.h"

int Automate_est_lexique(Automate lexique)
{
	if(lexique == NULL)
		return 0;
	if(! Automate_est_deterministe(lexique))
		return 0;
	return 1;
}

static int nb_transitions_vers_etat(const AVL cur, Etat etat)
{
	if(cur == NULL)
		return 0;
	return (cmp_etat(((Transition)(cur->valeur))->arrivee, etat) == 0)
	+ nb_transitions_vers_etat(cur->gauche, etat)
	+ nb_transitions_vers_etat(cur->droite, etat);
}

static int Automate_lexique_separer_action_transitions(Automate lexique, Mot nom_etat, const AVL cur, int indice)
{
	Mot nom_arrivee = NULL;
	Transition trans = NULL;
	if(cur == NULL)
		return 1;
	if((trans = (Transition)(cur->valeur)) == NULL)
		return 0;
	if((nom_arrivee = couple_chaine(trans->arrivee->nom, &indice, &Mot_chaine, &int_chaine)) == NULL)
		return 0;
	if(! Automate_existe_etat(lexique, nom_arrivee))
	{
		if(! Automate_ajouter_etat(lexique, nom_arrivee))
		{
			Element_free(TYPE_MOT)(&nom_arrivee);
			return 0;
		}
		if(Ensemble_trouver(lexique->etats_finaux, trans->arrivee))
		{
			if(! Automate_ajouter_etat_final(lexique, nom_arrivee))
			{
				Element_free(TYPE_MOT)(&nom_arrivee);
				return 0;
			}
		}
	}
	if(! Automate_existe_transition(lexique, nom_etat, trans->lettre, nom_arrivee))
	{
		if(! Automate_ajouter_transition(lexique, nom_etat, trans->lettre, nom_arrivee))
		{
			Element_free(TYPE_MOT)(&nom_arrivee);
			return 0;
		}
	}
	if(! Automate_lexique_separer_action_transitions(lexique, nom_etat, cur->gauche, indice)
	|| ! Automate_lexique_separer_action_transitions(lexique, nom_etat, cur->droite, indice)
	|| ! Automate_lexique_separer_action_transitions(lexique, nom_arrivee, trans->arrivee->suivants->elements, indice))
	{
		Element_free(TYPE_MOT)(&nom_arrivee);
		return 0;
	}
	Element_free(TYPE_MOT)(&nom_arrivee);
	return 1;
}

static int Automate_lexique_separer_action(Automate lexique, Transition trans, int indice)
{
	AVL cur = NULL;
	int retour;
	if((cur = AVL_creer_noeud(trans)) == NULL)
		return 0;
	retour = Automate_lexique_separer_action_transitions(lexique, trans->depart->nom, cur, indice);
	retour &= Automate_supprimer_transition(lexique, trans->depart->nom, trans->lettre, trans->arrivee->nom);
	AVL_free(&cur, &Void_free);
	return retour;
}

static int Automate_lexique_separer_sur_etat(Automate lexique, Etat etat, Mot mot, int indice)
{
	Transition suivant = NULL;
	AVL suivant_avl = NULL;
	Mot lettre = NULL;
	int taille_lettre = 0;
	
	if(*mot == '\0')
		return 1;
	if((lettre = premiere_lettre(mot)) == NULL)
		return 0;
	taille_lettre = strlen((char *)lettre);
	if((suivant_avl = AVL_trouver(etat->suivants->elements, lettre, cmp_transition_et_lettre)) == NULL)
	{
		Element_free(TYPE_MOT)(&lettre);
		return 1;
	}
	suivant = (Transition)(suivant_avl->valeur);
	if(nb_transitions_vers_etat(etat->suivants->elements, suivant->arrivee) > 1 || Ensemble_nombre_elements(suivant->arrivee->precedents) > 1)
	{
		++indice;
		if(! Automate_lexique_separer_action(lexique, suivant, indice))
		{
			Element_free(TYPE_MOT)(&lettre);
			return 0;
		}
		if((suivant_avl = AVL_trouver(etat->suivants->elements, lettre, cmp_transition_et_lettre)) == NULL)
		{
			Element_free(TYPE_MOT)(&lettre);
			return 0;
		}
		suivant = (Transition)(suivant_avl->valeur);
	}
	Element_free(TYPE_MOT)(&lettre);
	return Automate_lexique_separer_sur_etat(lexique, suivant->arrivee, mot + taille_lettre, indice);
}

int Automate_lexique_separer_pour_mot(Automate lexique, Mot mot)
{
	return Automate_lexique_separer_sur_etat(lexique, (Etat)(lexique->etats_init->elements->valeur), mot, 0);
}

static int Automate_lexique_ajouter_mot_action(Automate lexique, Etat etat, Mot mot, int indice)
{
	Transition suivant = NULL;
	AVL suivant_avl = NULL;
	Mot lettre = NULL;
	int taille_lettre = 0;
	
	if(etat == NULL)
		return 0;
	if(*mot == '\0')
	{
		if(! Automate_ajouter_etat_final(lexique, etat->nom))
			return 0;
		return 1;
	}
	if((lettre = premiere_lettre(mot)) == NULL)
		return 0;
	taille_lettre = strlen((char *)lettre);
	
	if((suivant_avl = AVL_trouver(etat->suivants->elements, lettre, cmp_transition_et_lettre)) == NULL)
	{
		Mot nom_arrivee = NULL;
		if((nom_arrivee = couple_chaine(&indice, "-1", &int_chaine, &Mot_chaine)) == NULL)
		{
			Element_free(TYPE_MOT)(&lettre);
			return 0;
		}
		if(! Automate_existe_etat(lexique, nom_arrivee))
		{
			if(! Automate_ajouter_etat(lexique, nom_arrivee))
			{
				Element_free(TYPE_MOT)(&lettre);
				Element_free(TYPE_MOT)(&nom_arrivee);
				return 0;
			}
		}
		if(! Automate_existe_transition(lexique, etat->nom, lettre, nom_arrivee))
		{
			if(! Automate_ajouter_transition(lexique, etat->nom, lettre, nom_arrivee))
			{
				Element_free(TYPE_MOT)(&lettre);
				Element_free(TYPE_MOT)(&nom_arrivee);
				return 0;
			}
		}
		Element_free(TYPE_MOT)(&nom_arrivee);
		if((suivant_avl = AVL_trouver(etat->suivants->elements, lettre, cmp_transition_et_lettre)) == NULL)
		{
			Element_free(TYPE_MOT)(&lettre);
			return 0;
		}
		++indice;
	}
	
	Element_free(TYPE_MOT)(&lettre);
	suivant = (Transition)(suivant_avl->valeur);
	return Automate_lexique_ajouter_mot_action(lexique, suivant->arrivee, mot + taille_lettre, indice);
}

static int Automate_lexique_ajouter_mot_action_prefixe(Automate lexique, Etat etat, Mot mot, Mot base, int indice)
{
	Transition suivant = NULL;
	AVL suivant_avl = NULL;
	Mot lettre = NULL;
	Mot_aux tmp;
	int taille_lettre = 0;
	
	if(etat == NULL)
		return 0;
	if(*mot == '\0')
	{
		if(! Automate_ajouter_etat_final(lexique, etat->nom))
			return 0;
		return 1;
	}
	if((lettre = premiere_lettre(mot)) == NULL)
		return 0;
	taille_lettre = strlen((char *)lettre);
	
	if((suivant_avl = AVL_trouver(etat->suivants->elements, lettre, cmp_transition_et_lettre)) == NULL)
	{
		Mot nom_arrivee = base;
		tmp = *(mot + taille_lettre);
		*(mot + taille_lettre) = (Mot_aux)'\0';
		if(! Automate_existe_etat(lexique, nom_arrivee))
		{
			if(! Automate_ajouter_etat(lexique, nom_arrivee))
			{
				Element_free(TYPE_MOT)(&lettre);
				return 0;
			}
		}
		if(! Automate_existe_transition(lexique, etat->nom, lettre, nom_arrivee))
		{
			if(! Automate_ajouter_transition(lexique, etat->nom, lettre, nom_arrivee))
			{
				Element_free(TYPE_MOT)(&lettre);
				return 0;
			}
		}
		*(mot + taille_lettre) = tmp;
		if((suivant_avl = AVL_trouver(etat->suivants->elements, lettre, cmp_transition_et_lettre)) == NULL)
		{
			Element_free(TYPE_MOT)(&lettre);
			return 0;
		}
		++indice;
	}
	
	Element_free(TYPE_MOT)(&lettre);
	suivant = (Transition)(suivant_avl->valeur);
	return Automate_lexique_ajouter_mot_action_prefixe(lexique, suivant->arrivee, mot + taille_lettre, base, indice);
}

int Automate_lexique_ajouter_mot(Automate * lexique, Mot mot)
{
	Automate nouv = NULL;
	if(lexique == NULL || *lexique == NULL)
		return 0;
	switch(Ensemble_nombre_elements((*lexique)->etats_init))
	{
		case 0:
		if((nouv = Automate_creer()) == NULL)
			return 0;
		if(! Automate_ajouter_etat(nouv, (Mot)"q0")
		|| ! Automate_ajouter_etat_initial(nouv, (Mot)"q0"))
		{
			Automate_free(&nouv);
			return 0;
		}
		break;
		
		case 1:
		if(! Automate_est_deterministe(*lexique))
			return 0;
		if(Automate_accepte_mot(*lexique, mot))
			return 1;
		if(! Automate_copier(&nouv, *lexique))
			return 0;
		if(! Automate_renommer_etats_par_nombre(nouv)
		|| ! Automate_lexique_separer_pour_mot(nouv, mot))
		{
			Automate_free(&nouv);
			return 0;
		}
		break;
		
		default:
		return 0;
		break;
	}
	
	if(! Automate_lexique_ajouter_mot_action(nouv, (Etat)(nouv->etats_init->elements->valeur), mot, 0)
	|| ! Automate_renommer_etats_par_nombre(nouv)
	|| ! Automate_completer(&nouv, (Mot)"-1")
	|| ! Automate_minimiser(&nouv)
	|| ! Automate_supprimer_etats_non_coaccessibles(nouv)
	|| ! Automate_renommer_etats_par_nombre(nouv))
	{
		Automate_free(&nouv);
		return 0;
	}
	Automate_free(lexique);
	*lexique = nouv;
	return 1;
}

int Automate_lexique_ajouter_mot_arbre(Automate * lexique, Mot mot)
{
	Automate nouv = NULL;
	if(lexique == NULL || *lexique == NULL)
		return 0;
	switch(Ensemble_nombre_elements((*lexique)->etats_init))
	{
		case 0:
		if((nouv = Automate_creer()) == NULL)
			return 0;
		if(! Automate_ajouter_etat(nouv, (Mot)"q0")
		|| ! Automate_ajouter_etat_initial(nouv, (Mot)"q0"))
		{
			Automate_free(&nouv);
			return 0;
		}
		break;
		
		case 1:
		if(! Automate_est_deterministe(*lexique))
			return 0;
		if(Automate_accepte_mot(*lexique, mot))
			return 1;
		if(! Automate_copier(&nouv, *lexique))
			return 0;
		if(! Automate_renommer_etats_par_nombre(nouv))
		{
			Automate_free(&nouv);
			return 0;
		}
		break;
		
		default:
		return 0;
		break;
	}
	
	if(! Automate_lexique_ajouter_mot_action(nouv, (Etat)(nouv->etats_init->elements->valeur), mot, 0)
	|| ! Automate_renommer_etats_par_nombre(nouv))
	{
		Automate_free(&nouv);
		return 0;
	}
	Automate_free(lexique);
	*lexique = nouv;
	return 1;
}

int Automate_lexique_ajouter_mot_arbre_prefixe(Automate * lexique, Mot mot)
{
	/*Automate nouv = NULL;*/
	if(lexique == NULL || *lexique == NULL)
		return 0;
	switch(Ensemble_nombre_elements((*lexique)->etats_init))
	{
		case 0:
		/*if((nouv = Automate_creer()) == NULL)
			return 0;*/
		if(! Automate_ajouter_etat(*lexique, (Mot)"")
		|| ! Automate_ajouter_etat_initial(*lexique, (Mot)""))
		{
			Automate_free(lexique);
			return 0;
		}
		break;
		
		case 1:
		/*if(! Automate_est_deterministe(*lexique))
			return 0;*/
		if(Automate_accepte_mot(*lexique, mot))
			return 1;
		/*if(! Automate_copier(&nouv, *lexique))
			return 0;*/
		break;
		
		default:
		return 0;
		break;
	}
	
	if(! Automate_lexique_ajouter_mot_action_prefixe(*lexique, (Etat)((*lexique)->etats_init->elements->valeur), mot, mot, 0))
	{
		/*Automate_free(lexique);*/
		return 0;
	}
	/*Automate_free(lexique);
	*lexique = nouv;*/
	return 1;
}

Etat etat_de_fin(Etat etat)
{
	if(etat == NULL)
		return NULL;
	if(etat->suivants->elements == NULL)
		return etat;
	return etat_de_fin(((Transition)(etat->suivants->elements->valeur))->arrivee);
}

static int envoyer_sur_fin(Automate lexique, Etat etat, Etat fin);

static int envoyer_sur_fin_transition(Automate lexique, const AVL cur, Etat fin)
{
	Transition trans = NULL;
	if(cur == NULL)
		return 1;
	if((trans = (Transition)(cur->valeur)) == NULL)
		return 0;
	if(Ensemble_nombre_elements(trans->arrivee->suivants) == 0)
	{
		if(! Automate_existe_transition(lexique, trans->depart->nom, trans->lettre, fin->nom))
		{
			if(! Automate_ajouter_transition(lexique, trans->depart->nom, trans->lettre, fin->nom))
				return 0;
		}
		else
			trans = NULL;
	}
	else
	{
		if(! envoyer_sur_fin(lexique, trans->arrivee, fin))
			return 0;
		trans = NULL;
	}
	
	if(! envoyer_sur_fin_transition(lexique, cur->gauche, fin))
		return 0;
	if(! envoyer_sur_fin_transition(lexique, cur->droite, fin))
		return 0;
	if(trans != NULL)
	{
		Etat arrivee = trans->arrivee;
		if(! Automate_supprimer_transition(lexique, trans->depart->nom, trans->lettre, trans->arrivee->nom))
			return 0;
		if(! Automate_supprimer_etat(lexique, arrivee->nom))
			return 0;
	}
	return 1;
}

static int envoyer_sur_fin(Automate lexique, Etat etat, Etat fin)
{
	if(etat == NULL)
		return 0;
	if(! envoyer_sur_fin_transition(lexique, etat->suivants->elements, fin))
		return 0;
	return 1;
}

static int Automate_joindre_etats_fin(Automate lexique)
{
	Etat fin = NULL;
	if(lexique == NULL)
		return 0;
	if(lexique->etats_init->elements == NULL)
		return 0;
	if((fin = etat_de_fin((Etat)(lexique->etats_init->elements->valeur))) == NULL)
		return 0;
	if(! envoyer_sur_fin(lexique, (Etat)(lexique->etats_init->elements->valeur), fin))
		return 0;
	return 1;
}

static int Automate_lexique_compacter_prefixe(Automate * lexique)
{
	if(lexique == NULL || *lexique == NULL)
		return 0;
	if(! Automate_joindre_etats_fin(*lexique))
		return 0;
	if(! Automate_renommer_etats_par_nombre(*lexique)
	/*|| ! Automate_completer(lexique, (Mot)"-1")
	|| ! Automate_minimiser(lexique)
	|| ! Automate_supprimer_etats_non_coaccessibles(*lexique)
	|| ! Automate_renommer_etats_par_nombre(*lexique)*/)
	{
		return 0;
	}
	return 1;
}

int Automate_lexique_compacter(Automate * lexique)
{
	Automate nouv = NULL;
	if(lexique == NULL || *lexique == NULL)
		return 0;
		
	if(Ensemble_nombre_elements((*lexique)->etats_init) != 1)
		return 0;
		
	if(! Automate_est_deterministe(*lexique))
		return 0;
		
	if(! Automate_copier(&nouv, *lexique))
		return 0;
		
	if(! Automate_renommer_etats_par_nombre(nouv)
	|| ! Automate_completer(&nouv, (Mot)"-1")
	|| ! Automate_minimiser(&nouv)
	|| ! Automate_supprimer_etats_non_coaccessibles(nouv)
	|| ! Automate_renommer_etats_par_nombre(nouv))
	{
		Automate_free(&nouv);
		return 0;
	}
	Automate_free(lexique);
	*lexique = nouv;
	return 1;
}

static int Automate_lexique_afficher_etat(Automate lexique, FILE * sortie, Etat etat, Mot * mot, int id, int * taille);

static int Automate_lexique_afficher_transitions(Automate lexique, FILE * sortie, const AVL cur, Mot * mot, int id, int * taille)
{
	Transition trans = NULL;
	int i;
	if(cur == NULL)
		return 1;
	if((trans = (Transition)(cur->valeur)) == NULL)
		return 0;
	if(! Automate_lexique_afficher_transitions(lexique, sortie, cur->gauche, mot, id, taille))
		return 0;
	while(id + strlen((char *)(trans->lettre)) >= *taille)
	{
		Mot tmp = NULL;
		if((tmp = (Mot)realloc(*mot, (*taille) * 2 * sizeof(Mot_aux))) == NULL)
			return 0;
		*mot = tmp;
		(*taille) *= 2;
	}
	for(i = 0; i < strlen((char *)(trans->lettre)); ++i)
		*((*mot) + id + i) = *(trans->lettre + i);
	*((*mot) + id + strlen((char *)(trans->lettre))) = '\0';
	if(! Automate_lexique_afficher_etat(lexique, sortie, trans->arrivee, mot, id + strlen((char *)(trans->lettre)), taille))
		return 0;
	if(! Automate_lexique_afficher_transitions(lexique, sortie, cur->droite, mot, id, taille))
		return 0;
	return 1;
}

static int Automate_lexique_afficher_etat(Automate lexique, FILE * sortie, Etat etat, Mot * mot, int id, int * taille)
{
	if(Ensemble_trouver(lexique->etats_finaux, etat))
		fprintf(sortie, "%s\n", *mot);
	if(! Automate_lexique_afficher_transitions(lexique, sortie, etat->suivants->elements, mot, id, taille))
		return 0;
	return 1;
}

int Automate_lexique_afficher(Automate lexique, FILE * sortie)
{
	Mot mot = NULL;
	int taille = 1;
	int retour;
	
	if(sortie == NULL)
		return 0;
	if(lexique == NULL)
		return 0;
	if(! Automate_est_lexique(lexique))
		return 0;
	if((mot = Mot_creer((Mot)"")) == NULL)
		return 0;
	retour = Automate_lexique_afficher_etat(lexique, sortie, (Etat)(lexique->etats_init->elements->valeur), &mot, 0, &taille);
	Element_free(TYPE_MOT)(&mot);
	return retour;
}

int Automate_lexique_lire(Automate * lexique, FILE * entree)
{
	int i = 0;
	int somme_lettres = 0;
	char mot[100];
	if(entree == NULL)
		return 0;
	while(fscanf(entree, "%99s", mot) == 1 && i < 100000)
	{
		printf("%d) Ajout de %s\n", i, mot);
		somme_lettres += strlen(mot);
		if(! Automate_lexique_ajouter_mot_arbre_prefixe(lexique, (Mot)mot))
			return 0;
		i ++;
	}
	printf("Somme = %d\n", somme_lettres);
	return Automate_lexique_compacter_prefixe(lexique);
}
