#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "automate_propriete.h"
#include "automate_edit.h"
#include "automate_utilisation.h"
#include "automate_aff_debug.h"

static int Automate_accepte_mot_transition(Automate automate, const AVL trans, Mot mot)
{
	Mot prem_mot = NULL;
	Mot prem_trans = NULL;
	
	if(trans == NULL)
		return 0;
	if(*mot == '\0')
		return 0;
	if((prem_mot = premiere_lettre(mot)) == NULL)
		return 0;
	if((prem_trans = premiere_lettre(((Transition)(trans->valeur))->lettre)) == NULL)
		return 0;
	/*printf("cmp (%c vs %c(%s))\n", (char)(*(((Transition)(trans->valeur))->lettre)), (char)(*mot), (char *)mot);*/
	if(cmp_mot(prem_trans, prem_mot) > 0)
	{
		Element_free(TYPE_MOT)(&prem_mot);
		Element_free(TYPE_MOT)(&prem_trans);
		return Automate_accepte_mot_transition(automate, trans->gauche, mot);
	}
	else if(cmp_mot(prem_trans, prem_mot) < 0)
	{
		Element_free(TYPE_MOT)(&prem_mot);
		Element_free(TYPE_MOT)(&prem_trans);
		return Automate_accepte_mot_transition(automate, trans->droite, mot);
	}
	
	if(*(mot + strlen((char *)prem_mot)) == '\0')
	{
		if(Ensemble_trouver(automate->etats_finaux, (Etat)(((Transition)(trans->valeur))->arrivee)))
		{
			Element_free(TYPE_MOT)(&prem_mot);
			Element_free(TYPE_MOT)(&prem_trans);
			return 1;
		}
	}
	else
	{
		if(Automate_accepte_mot_transition(automate, ((Etat)(((Transition)(trans->valeur))->arrivee))->suivants->elements, mot + strlen((char *)prem_mot)))
		{
			Element_free(TYPE_MOT)(&prem_mot);
			Element_free(TYPE_MOT)(&prem_trans);
			return 1;
		}
	}
	Element_free(TYPE_MOT)(&prem_mot);
	Element_free(TYPE_MOT)(&prem_trans);
	if(Automate_accepte_mot_transition(automate, trans->gauche, mot))
		return 1;
	if(Automate_accepte_mot_transition(automate, trans->droite, mot))
		return 1;
	
	return 0;
}

static int Automate_accepte_mot_depuis_etat(Automate automate, const AVL cur, Mot mot)
{
	if(cur == NULL)
		return 0;
	if(*mot == '\0' && Ensemble_trouver(automate->etats_finaux, ((Etat)(cur->valeur))))
		return 1;
	if(Automate_accepte_mot_transition(automate, ((Etat)(cur->valeur))->suivants->elements, mot))
		return 1;
	if(Automate_accepte_mot_depuis_etat(automate, cur->gauche, mot))
		return 1;
	if(Automate_accepte_mot_depuis_etat(automate, cur->droite, mot))
		return 1;
	return 0;
}

int Automate_accepte_mot(Automate automate, Mot mot)
{
	if(automate == NULL || mot == NULL)
		return 0;
	return Automate_accepte_mot_depuis_etat(automate, automate->etats_init->elements, mot);
}

static int Automate_completer_par_transitions(Automate automate, Mot nom_etat_depart, const AVL cur, Mot nom_etat_poubelle)
{
	if(cur == NULL)
		return 1;
	if(! Automate_ajouter_transition(automate, nom_etat_depart, (Mot)(cur->valeur), nom_etat_poubelle))
		return 0;
	if(! Automate_completer_par_transitions(automate, nom_etat_depart, cur->gauche, nom_etat_poubelle))
		return 0;
	if(! Automate_completer_par_transitions(automate, nom_etat_depart, cur->droite, nom_etat_poubelle))
		return 0;
	return 1;
}

static int Automate_completer_transitions(const AVL cur, Ensemble lettres)
{
	if(cur == NULL)
		return 1;
	if(Ensemble_trouver(lettres, ((Transition)(cur->valeur))->lettre))
	{
		if(! Ensemble_supprimer(lettres, ((Transition)(cur->valeur))->lettre))
			return 0;
	}
	if(! Automate_completer_transitions(cur->gauche, lettres))
		return 0;
	if(! Automate_completer_transitions(cur->droite, lettres))
		return 0;
	return 1;
}

static int Automate_completer_etat(Automate automate, const AVL cur, Mot nom_etat_poubelle)
{
	Ensemble lettres = NULL;
	int retour = 0;
	
	if(cur == NULL)
		return 1;
		
	if(! Ensemble_copier(&lettres, automate->alphabet, 1))
		return 0;
	if(Automate_completer_transitions(((Etat)(cur->valeur))->suivants->elements, lettres))
	{
		if(Automate_completer_par_transitions(automate, ((Etat)(cur->valeur))->nom, lettres->elements, nom_etat_poubelle))
			retour = 1;
	}
	Ensemble_free(&lettres);
	
	if(! Automate_completer_etat(automate, cur->gauche, nom_etat_poubelle))
		return 0;
	if(! Automate_completer_etat(automate, cur->droite, nom_etat_poubelle))
		return 0;
	
	return retour;
}

int Automate_completer(Automate * automate, Mot nom_etat_poubelle)
{
	Automate nouv = NULL;
	if(automate == NULL || *automate == NULL)
		return 0;
	if(Automate_est_complet(*automate))
		return 1;
	if(Automate_existe_etat(*automate, nom_etat_poubelle))
		return 0;
	if(! Automate_copier(&nouv, *automate))
		return 0;
	if(! Automate_ajouter_etat(nouv, nom_etat_poubelle)
	|| ! Automate_completer_etat(nouv, nouv->etats->elements, nom_etat_poubelle))
	{
		Automate_free(&nouv);
		return 0;
	}
	Automate_free(automate);
	*automate = nouv;
	return 1;
}

static int Automate_determiniser_transition(const AVL cur, Mot lettre, Ensemble arrivee)
{
	if(cur == NULL)
		return 1;
	if(cmp_transition_et_lettre(cur->valeur, lettre) > 0)
		return Automate_determiniser_transition(cur->gauche, lettre, arrivee);
	else if(cmp_transition_et_lettre(cur->valeur, lettre) < 0)
		return Automate_determiniser_transition(cur->droite, lettre, arrivee);
	
	if(! Ensemble_trouver(arrivee, ((Transition)(cur->valeur))->arrivee))
	{
		if(! Ensemble_ajouter(arrivee, ((Transition)(cur->valeur))->arrivee))
			return 0;
	}
	if(! Automate_determiniser_transition(cur->gauche, lettre, arrivee))
		return 0;
	if(! Automate_determiniser_transition(cur->droite, lettre, arrivee))
		return 0;
	return 1;
}

static int Automate_determiniser_etat(const AVL cur, Mot lettre, Ensemble arrivee)
{
	if(cur == NULL)
		return 1;
	if(! Automate_determiniser_transition(((Etat)(cur->valeur))->suivants->elements, lettre, arrivee))
		return 0;
	if(! Automate_determiniser_etat(cur->gauche, lettre, arrivee))
		return 0;
	if(! Automate_determiniser_etat(cur->droite, lettre, arrivee))
		return 0;
	return 1;
}

static int Automate_determiniser_ensemble(Automate deter, Automate base, Ensemble etats);

static int Automate_determiniser_alphabet(Automate deter, Automate base, Mot nom_etat, const AVL cur, Ensemble etats)
{
	Ensemble arrivee = NULL;
	Mot nom_arrivee = NULL;
	
	if(cur == NULL)
		return 1;
		
	if((arrivee = Ensemble_creer(TYPE_ETAT, 1)) == NULL)
		return 0;
		
	if(! Automate_determiniser_etat(etats->elements, (Mot)(cur->valeur), arrivee))
	{
		Ensemble_free(&arrivee);
		return 0;
	}
	
	if((nom_arrivee = Ensemble_chaine(arrivee)) == NULL
	|| ! Automate_determiniser_etat(etats->elements, (Mot)(cur->valeur), arrivee)
	|| ! Automate_determiniser_ensemble(deter, base, arrivee)
	|| ! Automate_ajouter_transition(deter, nom_etat, (Mot)(cur->valeur), nom_arrivee))
	{
		Element_free(TYPE_MOT)(&nom_arrivee);
		Ensemble_free(&arrivee);
		return 0;
	}
	
	Element_free(TYPE_MOT)(&nom_arrivee);
	Ensemble_free(&arrivee);
	if(! Automate_determiniser_alphabet(deter, base, nom_etat, cur->gauche, etats))
		return 0;
	if(! Automate_determiniser_alphabet(deter, base, nom_etat, cur->droite, etats))
		return 0;
	return 1;
}

static int Automate_ensemble_possede_etat_final(Automate automate, const AVL cur)
{
	if(cur == NULL)
		return 0;
	if(Ensemble_trouver(automate->etats_finaux, cur->valeur))
		return 1;
	return Automate_ensemble_possede_etat_final(automate, cur->gauche) || Automate_ensemble_possede_etat_final(automate, cur->droite);
}

static int Automate_determiniser_ensemble(Automate deter, Automate base, Ensemble etats)
{
	Mot nom_etat = NULL;
	
	if((nom_etat = Ensemble_chaine(etats)) == NULL)
		return 0;
		
	if(Automate_existe_etat(deter, nom_etat))
	{
		Element_free(TYPE_MOT)(&nom_etat);
		return 1;
	}
	
	if(! Automate_ajouter_etat(deter, nom_etat)
	|| ! Automate_determiniser_alphabet(deter, base, nom_etat, base->alphabet->elements, etats))
	{
		Element_free(TYPE_MOT)(&nom_etat);
		return 0;
	}
	
	if(Automate_ensemble_possede_etat_final(base, etats->elements))
	{
		if(! Automate_ajouter_etat_final(deter, nom_etat))
			return 0;
	}
	
	Element_free(TYPE_MOT)(&nom_etat);
	return 1;
}

int Automate_determiniser(Automate * automate)
{
	Automate deter = NULL;
	Mot nom_etat_init = NULL;
	
	if(automate == NULL || *automate == NULL)
		return 0;
	if(Automate_est_deterministe(*automate))
		return 1;
	if((deter = Automate_creer()) == NULL)
		return 0;
	
	if(! Automate_determiniser_ensemble(deter, *automate, (*automate)->etats_init)
	|| (nom_etat_init = Ensemble_chaine((*automate)->etats_init)) == NULL)
	{
		Automate_free(&deter);
		return 0;
	}
	if(! Automate_ajouter_etat_initial(deter, nom_etat_init))
	{
		Element_free(TYPE_MOT)(&nom_etat_init);
		Automate_free(&deter);
		return 0;
	}
	Element_free(TYPE_MOT)(&nom_etat_init);
	Automate_free(automate);
	*automate = deter;
	return 1;
}

int Automate_standardiser_transition(Automate automate, const AVL cur, Mot nom_etat_init)
{
	if(cur == NULL)
		return 1;
	if(! Automate_existe_transition(automate, nom_etat_init, ((Transition)(cur->valeur))->lettre, ((Transition)(cur->valeur))->arrivee->nom))
	{
		if(! Automate_ajouter_transition(automate, nom_etat_init, ((Transition)(cur->valeur))->lettre, ((Transition)(cur->valeur))->arrivee->nom))
			return 0;
	}
	if(! Automate_standardiser_transition(automate, cur->gauche, nom_etat_init))
		return 0;
	if(! Automate_standardiser_transition(automate, cur->droite, nom_etat_init))
		return 0;
	return 1;
}

int Automate_standardiser_etat(Automate automate, const AVL cur, Mot nom_etat_init)
{
	if(cur == NULL)
		return 1;
	if(! Automate_standardiser_transition(automate, ((Etat)(cur->valeur))->suivants->elements, nom_etat_init))
		return 0;
	if(! Automate_standardiser_etat(automate, cur->gauche, nom_etat_init))
		return 0;
	if(! Automate_standardiser_etat(automate, cur->droite, nom_etat_init))
		return 0;
	return 1;
}

int Automate_standardiser(Automate * automate, Mot nom_etat_init)
{
	int accepte_epsilon = 0;
	Automate nouv = NULL;
	if(automate == NULL || *automate == NULL)
		return 0;
	if(Automate_est_standard(*automate))
		return 1;
	if(Automate_existe_etat(*automate, nom_etat_init))
		return 0;
	if(! Automate_copier(&nouv, *automate))
		return 0;
	accepte_epsilon = Automate_accepte_mot(nouv, (Mot)"");
	if(! Automate_ajouter_etat(nouv, nom_etat_init)
	||! Automate_standardiser_etat(nouv, nouv->etats_init->elements, nom_etat_init))
	{
		Automate_free(&nouv);
		return 0;
	}
	Ensemble_vider(nouv->etats_init);
	if(! Automate_ajouter_etat_initial(nouv, nom_etat_init))
	{
		Automate_free(&nouv);
		return 0;
	}
	if(accepte_epsilon)
	{
		if(! Automate_ajouter_etat_final(nouv, nom_etat_init))
		{
			Automate_free(&nouv);
			return 0;
		}
	}
	if(! Automate_supprimer_etats_non_accessibles(nouv))
	{
		Automate_free(&nouv);
		return 0;
	}
	Automate_free(automate);
	*automate = nouv;
	return 1;
}

int Automate_union_standard(Automate * automate, Automate ajout, Mot prefixe1, Mot prefixe2, Mot nom_etat_init)
{
	Automate nouv = NULL;
	if(automate == NULL || *automate == NULL || ajout == NULL)
		return 0;
	if(! Automate_copier(&nouv, *automate))
		return 0;
	if(! Automate_ajouter_automate(nouv, ajout, prefixe1, prefixe2)
	|| ! Automate_standardiser(&nouv, nom_etat_init))
	{
		Automate_free(&nouv);
		return 0;
	}
	Automate_free(automate);
	*automate = nouv;
	return 1;
}

static int Automate_concatener_standard_transition(Automate dest, Mot final_nom, const AVL cur)
{
	Transition trans = NULL;
	if(cur == NULL)
		return 1;
	trans = (Transition)(cur->valeur);
	
	if(! Automate_existe_transition(dest, final_nom, trans->lettre, trans->arrivee->nom))
	{
		if(! Automate_ajouter_transition(dest, final_nom, trans->lettre, trans->arrivee->nom))
			return 0;
	}
	
	if(! Automate_concatener_standard_transition(dest, final_nom, cur->gauche))
		return 0;
	if(! Automate_concatener_standard_transition(dest, final_nom, cur->droite))
		return 0;
	return 1;
}

static int Automate_concatener_standard_init(Automate dest, Mot final_nom, const AVL cur)
{
	if(cur == NULL)
		return 1;
	if(! Automate_concatener_standard_transition(dest, final_nom, ((Etat)(cur->valeur))->suivants->elements))
		return 0;
	if(! Automate_concatener_standard_init(dest, final_nom, cur->gauche))
		return 0;
	if(! Automate_concatener_standard_init(dest, final_nom, cur->droite))
		return 0;
	return 1;
}

static int Automate_concatener_standard_finaux(Automate dest, const AVL cur, Ensemble init)
{
	if(cur == NULL)
		return 1;
	if(! Automate_concatener_standard_init(dest, ((Etat)(cur->valeur))->nom, init->elements))
		return 0;
	if(! Automate_concatener_standard_finaux(dest, cur->gauche, init))
		return 0;
	if(! Automate_concatener_standard_finaux(dest, cur->droite, init))
		return 0;
	return 1;
}

int Automate_concatener_standard(Automate * automate, Automate ajout, Mot prefixe1, Mot prefixe2)
{
	Ensemble finaux = NULL;
	Automate nouv = NULL;
	Automate autre = NULL;
	
	if(automate == NULL || *automate == NULL || ajout == NULL)
		return 0;
	if(! Automate_copier(&nouv, *automate)
	|| ! Automate_copier(&autre, ajout)
	|| ! Automate_renommer_etats_par_prefixe(nouv, prefixe1)
	|| ! Automate_renommer_etats_par_prefixe(autre, prefixe2)
	|| ! Automate_standardiser(&nouv, prefixe1)
	|| ! Automate_standardiser(&autre, prefixe2)
	|| ! Ensemble_copier(&finaux, nouv->etats_finaux, 1)
	|| ! Automate_ajouter_automate(nouv, autre, (Mot)"", (Mot)"")
	|| ! Automate_concatener_standard_finaux(nouv, finaux->elements, autre->etats_init))
	{
		Ensemble_free(&finaux);
		Automate_free(&nouv);
		Automate_free(&autre);
		return 0;
	}
	Ensemble_vider(nouv->etats_init);
	if(! Automate_ajouter_etat_initial(nouv, prefixe1)
	|| ! Automate_supprimer_etats_non_accessibles(nouv))
	{
		Ensemble_free(&finaux);
		Automate_free(&nouv);
		Automate_free(&autre);
		return 0;
	}
	if(! Automate_accepte_mot(autre, (Mot)""))
	{
		if(! Ensemble_moins(nouv->etats_finaux, finaux))
		{
			Ensemble_free(&finaux);
			Automate_free(&nouv);
			Automate_free(&autre);
			return 0;
		}
	}
	Ensemble_free(&finaux);
	Automate_free(&autre);
	Automate_free(automate);
	*automate = nouv;
	return 1;
}

int Automate_etoile_standard(Automate * automate, Mot nom_etat_init)
{
	Automate nouv = NULL;
	
	if(automate == NULL || *automate == NULL)
		return 0;
	if(! Automate_copier(&nouv, *automate)
	|| ! Automate_standardiser(&nouv, nom_etat_init)
	|| ! Automate_concatener_standard_finaux(nouv, nouv->etats_finaux->elements, nouv->etats_init))
	{
		Automate_free(&nouv);
		return 0;
	}
	if(! Automate_accepte_mot(nouv, (Mot)""))
	{
		if(! Automate_ajouter_etat_final(nouv, nom_etat_init))
		{
			Automate_free(&nouv);
			return 0;
		}
	}
	Automate_free(automate);
	*automate = nouv;
	return 1;
}

int Automate_complementaire(Automate * automate)
{
	Automate nouv = NULL;
	Ensemble finaux = NULL;
	
	if(automate == NULL || *automate == NULL)
		return 0;
	if(! Automate_copier(&nouv, *automate)
	|| ! Automate_determiniser(&nouv)
	|| ! Ensemble_copier(&finaux, nouv->etats_finaux, 1)
	|| ! Ensemble_union(nouv->etats_finaux, nouv->etats)
	|| ! Ensemble_moins(nouv->etats_finaux, finaux))
	{
		Ensemble_free(&finaux);
		Automate_free(&nouv);
		return 0;
	}
	Ensemble_free(&finaux);
	Automate_free(automate);
	*automate = nouv;
	return 1;
}

static int Automate_minimiser_etat(Automate automate, const AVL cur, Ensemble ** classes, int * taille)
{
	int i;
	
	if(cur == NULL)
		return 1;
	
	for(i = 0; i < *taille; ++i)
	{
		/*printf("ec %d/%d\n", i, *taille);*/
		if(Automate_etats_equivalents(automate, ((Etat)(cur->valeur))->nom, ((Etat)((*((*classes) + i))->elements->valeur))->nom))
		{
			if(! Ensemble_ajouter(*((*classes) + i), cur->valeur))
				return 0;
			break;
		}
	}
	if(i == *taille)
	{
		Ensemble * tmp = NULL;
		if((tmp = (Ensemble *)realloc(*classes, ((*taille) + 1) * sizeof(Ensemble))) == NULL)
			return 0;
		*classes = tmp;
		if((*((*classes) + (*taille)) = Ensemble_creer(TYPE_ETAT, 1)) == NULL)
			return 0;
		if(! Ensemble_ajouter(*((*classes) + (*taille)), cur->valeur))
			return 0;
		++(*taille);
	}
	
	if(! Automate_minimiser_etat(automate, cur->gauche, classes, taille))
		return 0;
	if(! Automate_minimiser_etat(automate, cur->droite, classes, taille))
		return 0;
	return 1;
}

static int Automate_construire_depuis_classes_transitions(Automate res, Mot depart_nom, const AVL cur, Ensemble * classes, int nb_classes)
{
	int i;
	Transition trans = NULL;
	Mot arrivee = NULL;
	
	if(cur == NULL)
		return 1;
	trans = (Transition)(cur->valeur);
	
	for(i = 0; i < nb_classes; ++i)
	{
		/*printf("tc %d/%d\n", i, nb_classes);*/
		if(Ensemble_trouver(*(classes + i), trans->arrivee))
		{
			if((arrivee = int_chaine(&i)) == NULL)
				return 0;
			if(! Automate_ajouter_transition(res, depart_nom, trans->lettre, arrivee))
			{
				Element_free(TYPE_MOT)(&arrivee);
				return 0;
			}
			Element_free(TYPE_MOT)(&arrivee);
			break;
		}
	}
	
	if(i == nb_classes)
		return 0;
		
	if(! Automate_construire_depuis_classes_transitions(res, depart_nom, cur->gauche, classes, nb_classes))
		return 0;
	if(! Automate_construire_depuis_classes_transitions(res, depart_nom, cur->droite, classes, nb_classes))
		return 0;
	return 1;
}

static int Automate_construire_depuis_classes(Automate * res, Automate base, Ensemble * classes, int nb_classes)
{
	int i;
	Mot depart = NULL;
	
	if((*res = Automate_creer()) == NULL)
		return 0;
	for(i = 0; i < nb_classes; ++i)
	{
		/*printf("c %d/%d\n", i, nb_classes);*/
		if((depart = int_chaine(&i)) == NULL)
		{
			Automate_free(res);
			return 0;
		}
		if(! Automate_ajouter_etat(*res, depart))
		{
			Element_free(TYPE_MOT)(&depart);
			Automate_free(res);
			return 0;
		}
		if(Ensemble_trouver(*(classes + i), base->etats_init->elements->valeur))
		{
			if(! Automate_ajouter_etat_initial(*res, depart))
			{
				Element_free(TYPE_MOT)(&depart);
				Automate_free(res);
				return 0;
			}
		}
		if(Ensemble_trouver(base->etats_finaux, (*(classes + i))->elements->valeur))
		{
			if(! Automate_ajouter_etat_final(*res, depart))
			{
				Element_free(TYPE_MOT)(&depart);
				Automate_free(res);
				return 0;
			}
		}
		Element_free(TYPE_MOT)(&depart);
	}
	
	for(i = 0; i < nb_classes; ++i)
	{
		if((depart = int_chaine(&i)) == NULL)
		{
			Automate_free(res);
			return 0;
		}
		if(! Automate_construire_depuis_classes_transitions(*res, depart, ((Etat)(*(classes + i))->elements->valeur)->suivants->elements, classes, nb_classes))
		{
			Element_free(TYPE_MOT)(&depart);
			Automate_free(res);
			return 0;
		}
		Element_free(TYPE_MOT)(&depart);
	}
	return 1;
}

int Automate_minimiser(Automate * automate)
{
	Automate nouv = NULL;
	Automate res = NULL;
	Ensemble * classes = NULL;
	int nb_classes = 0;
	int i;
	int retour = 1;
	
	if(automate == NULL || *automate == NULL)
		return 0;
	if(! Automate_copier(&nouv, *automate))
		return 0;
	if(! Automate_determiniser(&nouv))
	{
		Automate_free(&nouv);
		return 0;
	}
	if(! Automate_est_complet(nouv))
	{
		if(! Automate_renommer_etats_par_prefixe(nouv, (Mot)"q")
		|| ! Automate_completer(&nouv, (Mot)"q"))
		{
			Automate_free(&nouv);
			return 0;
		}
	}
	
	if(! Automate_minimiser_etat(nouv, nouv->etats->elements, &classes, &nb_classes))
		retour = 0;
	if(! Automate_construire_depuis_classes(&res, nouv, classes, nb_classes))
		retour = 0;
	if(classes != NULL)
	{
		for(i = 0; i < nb_classes; ++i)
		{
			Ensemble_free(classes + i);
		}
		free(classes);
	}
	if(! retour)
	{
		Automate_free(&nouv);
		Automate_free(&res);
		return 0;
	}
	Automate_free(&nouv);
	Automate_free(automate);
	*automate = res;
	return 1;
}

int Automate_reconnaitre_prefixes(Automate * automate)
{
	Automate nouv = NULL;
	
	if(automate == NULL || *automate == NULL)
		return 0;
	if(! Automate_copier(&nouv, *automate))
		return 0;
	if(! Automate_supprimer_etats_non_accessibles(nouv)
	|| ! Automate_supprimer_etats_non_coaccessibles(nouv)
	|| ! Ensemble_union(nouv->etats_finaux, nouv->etats))
	{
		Automate_free(&nouv);
		return 0;
	}
	Automate_free(automate);
	*automate = nouv;
	return 1;
}

int Automate_reconnaitre_suffixes(Automate * automate)
{
	Automate nouv = NULL;
	
	if(automate == NULL || *automate == NULL)
		return 0;
	if(! Automate_copier(&nouv, *automate))
		return 0;
	if(! Automate_supprimer_etats_non_accessibles(nouv)
	|| ! Automate_supprimer_etats_non_coaccessibles(nouv)
	|| ! Ensemble_union(nouv->etats_init, nouv->etats))
	{
		Automate_free(&nouv);
		return 0;
	}
	Automate_free(automate);
	*automate = nouv;
	return 1;
}

int Automate_reconnaitre_facteurs(Automate * automate)
{
	Automate nouv = NULL;
	
	if(automate == NULL || *automate == NULL)
		return 0;
	if(! Automate_copier(&nouv, *automate))
		return 0;
	if(! Automate_supprimer_etats_non_accessibles(nouv)
	|| ! Automate_supprimer_etats_non_coaccessibles(nouv)
	|| ! Ensemble_union(nouv->etats_init, nouv->etats)
	|| ! Ensemble_union(nouv->etats_finaux, nouv->etats))
	{
		Automate_free(&nouv);
		return 0;
	}
	Automate_free(automate);
	*automate = nouv;
	return 1;
}

static int Automate_produit_etats(Automate res, Automate prem, Automate sec, Etat prem_etat, Etat sec_etat, int (*est_final)(Automate, Automate, Etat, Etat));

static int Automate_produit_transitions_aux(Automate res, Automate prem, Automate sec, Mot nom_etat, Transition prem_trans, const AVL cur, int (*est_final)(Automate, Automate, Etat, Etat))
{
	Transition sec_trans = NULL;
	Mot couple = NULL;
	if(cur == NULL)
		return 1;
	if((sec_trans = (Transition)(cur->valeur)) == NULL)
		return 0;
	if(cmp_mot(prem_trans->lettre, sec_trans->lettre) != 0)
	{
		if(! Automate_produit_transitions_aux(res, prem, sec, nom_etat, prem_trans, cur->gauche, est_final))
			return 0;
		if(! Automate_produit_transitions_aux(res, prem, sec, nom_etat, prem_trans, cur->droite, est_final))
			return 0;
		return 1;
	}
	if(! Automate_produit_etats(res, prem, sec, prem_trans->arrivee, sec_trans->arrivee, est_final))
		return 0;
		
	if(! Automate_produit_transitions_aux(res, prem, sec, nom_etat, prem_trans, cur->gauche, est_final))
		return 0;
	if(! Automate_produit_transitions_aux(res, prem, sec, nom_etat, prem_trans, cur->droite, est_final))
		return 0;
		
	if((couple = couple_chaine(prem_trans->arrivee->nom, sec_trans->arrivee->nom, &Mot_chaine, &Mot_chaine)) == NULL)
		return 0;
	if(! Automate_existe_transition(res, nom_etat, prem_trans->lettre, couple))
	{
		if(! Automate_ajouter_transition(res, nom_etat, prem_trans->lettre, couple))
		{
			Element_free(TYPE_MOT)(&couple);
			return 0;
		}
	}
	Element_free(TYPE_MOT)(&couple);
	return 1;
}

static int Automate_produit_transitions(Automate res, Automate prem, Automate sec, Mot nom_etat, const AVL cur, Etat sec_etat, int (*est_final)(Automate, Automate, Etat, Etat))
{
	if(cur == NULL)
		return 1;
	if(! Automate_produit_transitions_aux(res, prem, sec, nom_etat, (Transition)(cur->valeur), sec_etat->suivants->elements, est_final))
		return 0;
	if(! Automate_produit_transitions(res, prem, sec, nom_etat, cur->gauche, sec_etat, est_final))
		return 0;
	if(! Automate_produit_transitions(res, prem, sec, nom_etat, cur->droite, sec_etat, est_final))
		return 0;
	return 1;
}

static int Automate_produit_etats(Automate res, Automate prem, Automate sec, Etat prem_etat, Etat sec_etat, int (*est_final)(Automate, Automate, Etat, Etat))
{
	Mot couple = NULL;
	if(prem_etat == NULL || sec_etat == NULL)
		return 0;
	if((couple = couple_chaine(prem_etat->nom, sec_etat->nom, &Mot_chaine, &Mot_chaine)) == NULL)
		return 0;
	if(Automate_existe_etat(res, couple))
	{
		Element_free(TYPE_MOT)(&couple);
		return 1;
	}
	if(! Automate_ajouter_etat(res, couple))
	{
		Element_free(TYPE_MOT)(&couple);
		return 0;
	}
	if(est_final != NULL && (*est_final)(prem, sec, prem_etat, sec_etat))
	{
		if(! Automate_ajouter_etat_final(res, couple))
		{
			Element_free(TYPE_MOT)(&couple);
			return 0;
		}
	}
	if(! Automate_produit_transitions(res, prem, sec, couple, prem_etat->suivants->elements, sec_etat, est_final))
	{
		Element_free(TYPE_MOT)(&couple);
		return 0;
	}
	Element_free(TYPE_MOT)(&couple);
	return 1;
}

static int Automate_produit_init_aux(Automate res, Automate prem, Automate sec, Etat prem_etat, const AVL sec_cur, int (*est_final)(Automate, Automate, Etat, Etat))
{
	Mot couple = NULL;
	Etat sec_etat = NULL;
	if(sec_cur == NULL)
		return 1;
	sec_etat = (Etat)(sec_cur->valeur);
	if(! Automate_produit_etats(res, prem, sec, prem_etat, sec_etat, est_final))
		return 0;
	if(! Automate_produit_init_aux(res, prem, sec, prem_etat, sec_cur->gauche, est_final))
		return 0;
	if(! Automate_produit_init_aux(res, prem, sec, prem_etat, sec_cur->droite, est_final))
		return 0;
	if((couple = couple_chaine(prem_etat->nom, sec_etat->nom, &Mot_chaine, &Mot_chaine)) == NULL)
		return 0;
	if(! Automate_ajouter_etat_initial(res, couple))
	{
		Element_free(TYPE_MOT)(&couple);
		return 0;
	}
	Element_free(TYPE_MOT)(&couple);
	return 1;
}

static int Automate_produit_init(Automate res, Automate prem, Automate sec, const AVL prem_cur, int (*est_final)(Automate, Automate, Etat, Etat))
{
	if(prem_cur == NULL)
		return 1;
	if(! Automate_produit_init_aux(res, prem, sec, (Etat)(prem_cur->valeur), sec->etats_init->elements, est_final))
		return 0;
	if(! Automate_produit_init(res, prem, sec, prem_cur->gauche, est_final))
		return 0;
	if(! Automate_produit_init(res, prem, sec, prem_cur->droite, est_final))
		return 0;
	return 1;
}

int Automate_produit(Automate * automate, Automate ajout, int (*est_final)(Automate, Automate, Etat, Etat))
{
	Automate nouv = NULL;
	Automate prem = NULL;
	Automate sec = NULL;
	
	if(automate == NULL || *automate == NULL || ajout == NULL)
		return 0;
		
	if(est_final == &Automate_produit_union)
	{
		if(! Automate_copier(&prem, *automate)
		|| ! Automate_copier(&sec, ajout)
		|| ! Ensemble_union(prem->alphabet, sec->alphabet)
		|| ! Ensemble_union(sec->alphabet, prem->alphabet))
		{
			Automate_free(&prem);
			Automate_free(&sec);
			return 0;
		}
		
		if(! Automate_est_complet(prem))
		{
			if(! Automate_renommer_etats_par_prefixe(prem, (Mot)"q")
			|| ! Automate_completer(&prem, (Mot)"q"))
			{
				Automate_free(&prem);
				Automate_free(&sec);
				return 0;
			}
		}
		if(! Automate_est_complet(sec))
		{
			if(! Automate_renommer_etats_par_prefixe(sec, (Mot)"q")
			|| ! Automate_completer(&sec, (Mot)"q"))
			{
				Automate_free(&prem);
				Automate_free(&sec);
				return 0;
			}
		}
	}
	else if(est_final == &Automate_produit_difference)
	{
		if(! Automate_copier(&sec, ajout)
		|| ! Ensemble_union(sec->alphabet, (*automate)->alphabet))
		{
			Automate_free(&sec);
			return 0;
		}
		if(! Automate_est_deterministe(sec))
		{
			if(! Automate_determiniser(&sec))
			{
				Automate_free(&sec);
				return 0;
			}
		}
		else if(! Automate_est_complet(sec))
		{
			if(! Automate_renommer_etats_par_prefixe(sec, (Mot)"q")
			|| ! Automate_completer(&sec, (Mot)"q"))
			{
				Automate_free(&sec);
				return 0;
			}
		}
	}
	if(prem == NULL)
		prem = *automate;
	if(sec == NULL)
		sec = ajout;
	
	if((nouv = Automate_creer()) == NULL)
	{
		if(prem != *automate)
			Automate_free(&prem);
		if(sec != ajout)
			Automate_free(&sec);
		return 0;
	}
	
	if(! Automate_produit_init(nouv, prem, sec, prem->etats_init->elements, est_final))
	{
		Automate_free(&nouv);
		if(prem != *automate)
			Automate_free(&prem);
		if(sec != ajout)
			Automate_free(&sec);
		return 0;
	}
	
	if(prem != *automate)
		Automate_free(&prem);
	if(sec != ajout)
		Automate_free(&sec);
	
	Automate_free(automate);
	*automate = nouv;
	return 1;
}

int Automate_produit_union(Automate prem, Automate sec, Etat prem_etat, Etat sec_etat)
{
	return Ensemble_trouver(prem->etats_finaux, prem_etat) || Ensemble_trouver(sec->etats_finaux, sec_etat);
}

int Automate_produit_intersection(Automate prem, Automate sec, Etat prem_etat, Etat sec_etat)
{
	return Ensemble_trouver(prem->etats_finaux, prem_etat) && Ensemble_trouver(sec->etats_finaux, sec_etat);
}

int Automate_produit_difference(Automate prem, Automate sec, Etat prem_etat, Etat sec_etat)
{
	return Ensemble_trouver(prem->etats_finaux, prem_etat) && (! Ensemble_trouver(sec->etats_finaux, sec_etat));
}
