#ifndef DEF_HEADER_ELEMENTS
#define DEF_HEADER_ELEMENTS

#include "AVL.h"

#define CMP_NULL_VERIF(a, b) 	if(((a) == NULL) ^ ((b) == NULL))	\
				{					\
					if((a) == NULL)			\
						return -1;		\
					else				\
						return 1;		\
				}					\
				if((a) == NULL)				\
					return 0;
					

typedef enum
{
	TYPE_ENTIER,
	TYPE_MOT, /*Utilisé principalement comme lettre mais permet aussi de gérer une lettre en UTF8 ou un mot*/
	TYPE_ETAT,
	TYPE_TRANSITION
} Element_Type;

typedef struct _Ensemble Ensemble_aux, *Ensemble;
struct _Ensemble
{
	Element_Type type;
	AVL elements;
	unsigned int liens : 1; /*Indique s'il contient des liens ou des valeurs allouées dynamiquement*/
};

typedef unsigned char * Mot_aux, *Mot;

typedef struct _Etat Etat_aux, *Etat;
struct _Etat
{
	Mot nom;
	Ensemble suivants;/*liens vers transitions*/
	Ensemble precedents;/*liens vers transitions*/
};

typedef struct _Transition Transition_aux, *Transition;
struct _Transition
{
	Mot lettre;
	Etat depart;
	Etat arrivee;
};

/* Creation */

Mot Mot_creer(Mot chaine);

Etat Etat_creer(Mot nom);

Transition Transition_creer(Etat depart, Mot lettre, Etat arrivee);

Ensemble Ensemble_creer(Element_Type type, unsigned int liens);

/* Libération */

void (*Element_Ensemble_free(const Ensemble ens))(void *);

void (*Element_free(Element_Type type))(void *);

void Void_free(void * elem); /* Utile pour gérer les liens */

void Scalaire_free(void * elem);

void Etat_free(void * etat);

void Transition_free(void * transition);

void Ensemble_free(Ensemble * ensemble);

/* Comparaison */

int (*cmp_type(Element_Type type))(const void *, const void *);

int cmp_ptr(const void * a, const void * b);

int cmp_int(const void * a, const void * b);

int cmp_mot(const void * a, const void * b);

int cmp_etat(const void * a, const void * b);

int cmp_transition(const void * a, const void * b);

int cmp_etat_et_nom(const void * a, const void * b);

int cmp_transition_et_lettre(const void * a, const void * b);

int cmp_transition_et_depart(const void * a, const void * b);

int cmp_transition_et_arrivee(const void * a, const void * b);

/* Edition */

int Ensemble_ajouter(Ensemble ens, const void * valeur);

int Ensemble_trouver(Ensemble ens, const void * valeur);

int Ensemble_trouver_par(Ensemble ens, const void * valeur, Comparaison cmp);

int Ensemble_supprimer(Ensemble ens, const void * valeur);

void Ensemble_vider(Ensemble ens);

Etat Ensemble_trouver_etat_nom(Ensemble ens, Mot nom);

Transition Ensemble_trouver_transition(Ensemble ens, Etat depart, Mot lettre, Etat arrivee);

/* Copie */

int (*Element_copier(const Ensemble ens))(void **, const void *);

int int_copier(void ** dest, const void * source);

int Mot_copier(void ** dest, const void * source);

int Etat_copier(void ** dest, const void * source);

int Transition_copier(void ** dest, const void * source);

int lien_copier(void ** dest, const void * source);

int Ensemble_copier(Ensemble * dest, const Ensemble source, unsigned int liens);

int AVL_copier(AVL * dest, const AVL source, int copier(void **, const void *), void (*free_element)(void *));

/* Autres */

int Ensemble_union(Ensemble ens, Ensemble ajout);

int Ensemble_moins(Ensemble ens, Ensemble supp);

int Ensemble_egaux(Ensemble ens, Ensemble autre);

int Ensemble_nombre_elements(const Ensemble ens);

int Ensemble_restructurer(Ensemble ens);

#endif
